import logging

from yupi import DiffMethod, Trajectory, WindowType
from yupi.core import featurizers

from pactus.dataset import Dataset
from pactus.models import Evaluation, EvaluationComparison

__version__ = "0.1.1a2"

__all__ = [
    "Dataset",
    "Evaluation",
    "EvaluationComparison",
    "featurizers",
]

Trajectory.global_diff_method(
    method=DiffMethod.LINEAR_DIFF, window_type=WindowType.FORWARD
)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s]: %(message)s"
)



import json
from pathlib import Path
from string import Template
from typing import Any, List

import numpy as np
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    precision_recall_fscore_support,
)

from pactus import config
from pactus.dataset import Data
from pactus.dataset._utils import _get_path

MAIN_SEP = "="
SUB_SEP = "-"

LATEX_CM_ROW_TEMPLATE = Template(
    r"""            & $cls_name & $cls_vals & $cls_prec \\
"""
)

LATEX_CM_TEMPLATE = Template(
    r"""
\begin{figure}[ht]
\caption{$caption}
\vspace{2mm}
\centering
    \begin{tabular}{cc${c_cols}c}
        \toprule
        \multicolumn{2}{c}{\multirow{2}[4]{*}{\bf $model_name}} &
        \multicolumn{$cls_count}{c}{\bf Actual} &
        \multirow{2}[4]{*}{\bf Precision} \\
        & \cline{2-$c_line_top}
        & & $cls_head & \\
        \midrule
        \multirow{$cls_count}{*}{\bf Predicted}
$cls_rows            \midrule
        \multicolumn{2}{c}{\bf Recall} & $recalls \\
        \bottomrule
    \end{tabular}
\end{figure}
"""
)


class Evaluation:
    def __init__(
        self,
        model_summary: dict,
        data: Data,
        predictions: List[Any],
    ):
        self.dataset = data.dataset
        self.trajs = data.trajs
        self.y_true = data.labels
        self.y_pred = predictions
        self.model_summary = model_summary
        self.classes = list(set(self.y_true))

        self._confusion_matrix = confusion_matrix(
            self.y_true, self.y_pred, labels=self.classes
        ).T
        pre, rec, f_sc, sup = precision_recall_fscore_support(
            self.y_true, self.y_pred, labels=self.classes
        )
        self.precision = np.asarray(pre)
        self.recall = np.asarray(rec)
        self.f_score = np.asarray(f_sc)
        self.support = np.asarray(sup)

        self.acc_overall = accuracy_score(self.y_true, self.y_pred, normalize=True)
        self.f1_score = f1_score(self.y_true, self.y_pred, average="weighted")

    def _conf_matrix_perc(self) -> np.ndarray:
        c_matrix = self._confusion_matrix.astype("float")
        for i in range(c_matrix.shape[0]):
            c_matrix[:, i] /= c_matrix[:, i].sum()
        return c_matrix

    def _show_confusion_matrix(self):
        """Show the confusion matrix."""
        print("\nConfusion matrix:\n")

        # Normalize the confusion matrix by columns
        c_matrix = self._conf_matrix_perc()

        classes = list(self.classes) + ["precision"]
        col_width = 12

        print(*[f"{c:<12}".format() for c in classes], sep="")
        print(MAIN_SEP * col_width * (len(classes)))
        for i, row in enumerate(c_matrix):
            row = np.append(row, self.precision[i])
            print(*[f"{round(c * 100, 2):<12}" for c in row], sep="")
        print(SUB_SEP * col_width * (len(classes)))
        print(
            *[f"{round(rc * 100, 2):<12}" for rc in self.recall],
            sep="",
        )

    def show(self):
        """Show the evaluation results."""
        self._show_confusion_matrix()

    def save(self, file_name: str) -> Path:
        """Save the evaluation to a file.

        Parameters
        ----------
        file_name : str
            The name of the file to save the evaluation to. It
            must end with '.json'.

        Returns
        -------
        Path
            The path to the saved file.
        """

        if not file_name.endswith(".json"):
            raise ValueError("file_name extension must be '.json'")

        data = {
            "indices": [
                int(traj.traj_id) for traj in self.trajs if traj.traj_id is not None
            ],
            "predictions": self.y_pred,
            "model_summary": self.model_summary,
        }

        assert len(data["indices"]) == len(data["predictions"])

        file_path = _get_path(config.DS_EVALS_DIR, self.dataset.name) / file_name

        with open(file_path, "w", encoding="utf-8") as data_fd:
            json.dump(data, data_fd, ensure_ascii=False, indent=4)

        return file_path

    def to_markdown(self) -> str:
        """Evaluation summary in markdown style."""
        summary = self.model_summary.copy()
        model_name = summary.pop("name")
        ans = "# Evaluation results\n\n"
        ans += f"**Dataset:** {self.dataset.name} \\\n"
        ans += f"**Model:** {model_name}\n"
        ans += "\n## Model Summary\n\n"
        for param, val in summary.items():
            ans += f"- `{param} = {val}`\n"
        ans += "\n## Confusion Matrix\n\n"

        c_matrix = self._conf_matrix_perc()

        head = (
            "| Predicted \\ Actual | "
            + " | ".join(map(str, self.classes))
            + " | Precision |\n"
        )
        sep = "| :--: | " + " | ".join([":--:" for _ in self.classes]) + " | :--: |\n"
        body = ""
        for i, row in enumerate(c_matrix):
            row = np.append(row, self.precision[i])
            str_row = [str(round(c * 100, 2)) for c in row]
            str_row[i] = f"**{str_row[i]}**"
            body += f"| **{self.classes[i]}** | " + " | ".join(str_row) + " |\n"
        recall = (
            "| **Recall** | "
            + " | ".join([str(round(rc * 100, 2)) for rc in self.recall])
            + " |\n"
        )
        ans += head + sep + body + recall
        return ans

    def to_latex(self) -> str:
        """Evaluation summary in latex style."""
        summary = self.model_summary.copy()
        model_name = summary.pop("name")
        model_name = " ".join([val.title() for val in model_name.split("_")])
        ans = cls_rows = ""
        c_matrix = self._conf_matrix_perc()
        classes = [cls.replace("_", "\\_") for cls in self.classes]

        for i, row in enumerate(c_matrix):
            str_row = [f"{str(round(c * 100, 2))} \\%" for c in row]
            str_row[i] = r"\textbf{" + str_row[i] + "}"
            row = LATEX_CM_ROW_TEMPLATE.substitute(
                cls_name=classes[i],
                cls_vals=" & ".join(str_row),
                cls_prec=f"{str(round(self.precision[i] * 100, 2))} \\%",
            )
            cls_rows += row

        ans += LATEX_CM_TEMPLATE.substitute(
            caption=f"Confusion matrix for {model_name}. Dataset: {self.dataset.name}",
            model_name=model_name,
            c_cols="c" * len(classes),
            c_line_top=str(len(classes) + 1),
            cls_count=len(classes),
            cls_head=" & ".join(classes),
            cls_rows=cls_rows,
            recalls=" & ".join(
                [f"{str(round(rc * 100, 2))} \\%" for rc in self.precision]
            ),
        )
        return ans



import logging
from pathlib import Path
from typing import Any, List, Tuple, Union

import numpy as np
from sklearn.model_selection import KFold
from sklearn.preprocessing import LabelEncoder
from tensorflow import keras

import pactus.config as cfg
from pactus.dataset import Data
from pactus.models import Model
from pactus.models.transformer import build_model

NAME = "transformer_model"
DEFAULT_OPTIMIZER = keras.optimizers.Adam(learning_rate=1e-2)
DEFAULT_CALLBACKS = [
    keras.callbacks.EarlyStopping(patience=10, restore_best_weights=True)
]


class TransformerModel(Model):
    """Implementation of a Transformer model."""

    def __init__(
        self,
        head_size: int = 256,
        num_heads: int = 1,
        ff_dim: int = 4,
        num_transformer_blocks: int = 2,
        mlp_units: Union[List[int], None] = None,
        mlp_droput: float = 0.4,
        droput: float = 0.25,
        loss="categorical_crossentropy",
        optimizer=None,
        metrics=None,
        max_traj_len: int = -1,
        skip_long_trajs: bool = False,
        mask_value=cfg.MASK_VALUE,
    ):
        super().__init__(NAME)
        self.head_size = head_size
        self.num_heads = num_heads
        self.ff_dim = ff_dim
        self.num_transformer_blocks = num_transformer_blocks
        self.mlp_units = [128] if mlp_units is None else mlp_units
        self.mlp_dropout = mlp_droput
        self.dropout = droput
        self.model: keras.Model
        self.loss = loss
        self.optimizer = DEFAULT_OPTIMIZER if optimizer is None else optimizer
        self.metrics = ["accuracy"] if metrics is None else metrics
        self.max_traj_len = max_traj_len
        self.skip_long_trajs = skip_long_trajs
        self.mask_value = mask_value
        self.set_summary(
            head_size=self.head_size,
            num_heads=self.num_heads,
            ff_dim=self.ff_dim,
            num_transformer_blocks=self.num_transformer_blocks,
            mlp_units=self.mlp_units,
            mlp_dropout=self.mlp_dropout,
            dropout=self.dropout,
            loss=self.loss,
            optimizer=self.optimizer,
            metrics=self.metrics,
            max_traj_len=self.max_traj_len,
            skip_long_trajs=self.skip_long_trajs,
        )

    def train(
        self,
        data: Data,
        cross_validation: int = 0,
        epochs: int = 10,
        validation_split: float = 0.2,
        batch_size: int = 32,
        callbacks: Union[list, None] = None,
        checkpoint: Union[keras.callbacks.ModelCheckpoint, None] = None,
    ):
        self.set_summary(
            cross_validation=cross_validation,
            epochs=epochs,
            validation_split=validation_split,
            batch_size=batch_size,
        )
        x_train, y_train = self._get_input_data(data)
        n_classes = len(data.dataset.classes)
        input_shape = x_train.shape[1:]
        callbacks = DEFAULT_CALLBACKS if callbacks is None else callbacks
        model_path = None
        if checkpoint is not None:
            callbacks.append(checkpoint)
            if Path(checkpoint.filepath).exists():
                logging.info("Loading model from checkpoint %s", checkpoint.filepath)
                model_path = checkpoint.filepath

        if cross_validation == 0:
            model = (
                self._get_model(
                    n_classes,
                    input_shape,
                    mask=self.mask_value,
                )
                if model_path is None
                else keras.models.load_model(model_path)
            )
            model.fit(
                x_train,
                y_train,
                validation_split=validation_split,
                epochs=epochs,
                batch_size=batch_size,
                callbacks=callbacks,
            )
            self.model = model
        else:
            assert cross_validation > 1, "cross_validation must be greater than 1"
            kfold = KFold(n_splits=cross_validation, shuffle=True)

            best_acc = -1
            fold_no = 1
            for train_idxs, test_idxs in kfold.split(x_train, y_train):
                x_train_fold = x_train[train_idxs]
                y_train_fold = y_train[train_idxs]
                model = self._get_model(
                    n_classes,
                    input_shape,
                    mask=self.mask_value,
                )
                model.fit(
                    x_train_fold,
                    y_train_fold,
                    validation_split=validation_split,
                    epochs=epochs,
                    batch_size=batch_size,
                    callbacks=callbacks,
                )

                x_test_fold = x_train[test_idxs]
                y_test_fold = y_train[test_idxs]
                scores = model.evaluate(x_test_fold, y_test_fold, verbose=2)
                acc = scores[1]
                loss = scores[0]

                logging.info("Fold %d: Loss: %f, Accuracy: %f", fold_no, loss, acc)

                if acc > best_acc:
                    self.model = model
                fold_no += 1

    def predict(self, data: Data) -> List[Any]:
        x_data, _ = self._get_input_data(data)
        return self.model.predict(x_data)

    def _get_model(
        self, n_classes: int, input_shape: tuple, mask: Any = None
    ) -> keras.Model:
        model = build_model(
            n_classes,
            input_shape,
            head_size=self.head_size,
            num_heads=self.num_heads,
            ff_dim=self.ff_dim,
            num_transformer_blocks=self.num_transformer_blocks,
            mlp_units=self.mlp_units,
            mlp_dropout=self.mlp_dropout,
            dropout=self.dropout,
            mask=mask,
        )
        model.compile(
            loss=self.loss,
            optimizer=self.optimizer,
            metrics=self.metrics,
        )
        return model

    def _get_input_data(self, data: Data) -> Tuple[np.ndarray, np.ndarray]:
        """
        Process all the data and returns a x_data, y_data, mask readable
        by the transformer
        """
        y_data = self._encode_labels(data)
        x_data = self._extract_raw_data(data)
        x_data = self._reshape_input(x_data)
        return x_data, y_data

    def _encode_labels(self, data: Data) -> np.ndarray:
        """Encode the labels"""
        encoder = LabelEncoder()
        encoder.fit(data.dataset.labels)
        encoded_labels = encoder.transform(data.labels)
        assert isinstance(encoded_labels, np.ndarray)

        classes = np.zeros((len(encoded_labels), len(encoder.classes_)))
        for i, label in enumerate(encoded_labels):
            classes[i][label] = 1
        return classes

    def _extract_raw_data(self, data: Data) -> np.ndarray:
        """Extracts the raw data from the yupi trajectories"""
        trajs = data.trajs
        max_len = np.max([len(traj) for traj in data.dataset.trajs])
        if self.max_traj_len > 0:
            max_len = self.max_traj_len
        raw_data = [np.hstack((traj.r, np.reshape(traj.t, (-1, 1)))) for traj in trajs]
        if self.skip_long_trajs:
            raw_data = [traj for traj in raw_data if traj.shape[0] <= max_len]
        assert len(raw_data) > 0, "No trajectories to train on"
        all_raw_data = np.zeros((len(raw_data), max_len, 3))
        for i, traj in enumerate(raw_data):
            traj = traj[:max_len]
            all_raw_data[i, :, :] = self.mask_value
            all_raw_data[i, : traj.shape[0]] = traj
        return all_raw_data

    def _reshape_input(self, x_data: np.ndarray) -> np.ndarray:
        """Reshapes the input data to be compatible with the transformer."""
        return x_data.reshape((x_data.shape[0], x_data.shape[1], x_data.shape[2], 1))



from __future__ import annotations

import json
import logging
from collections import Counter
from pathlib import Path
from typing import Any, Callable, List, Tuple, Union

from git.cmd import Git
from sklearn.model_selection import train_test_split
from yupi import Trajectory
from yupi.core import JSONSerializer

from pactus import config
from pactus.dataset._utils import _get_path, download_dataset

REPO_URL = "https://github.com/yupidevs/trajectory-datasets"


class Data:
    """
    Structure that groups the trajectories and labels along with some
    useful methods to work with the set of them.
    """

    def __init__(
        self, dataset: Dataset, trajs: List[Trajectory], labels: List[Any]
    ) -> None:
        self.dataset = dataset
        self.trajs = trajs
        self.labels = labels
        self.label_counts = Counter(labels)

    @property
    def classes(self) -> List[Any]:
        """Classes present in the dataset."""
        return list(self.label_counts.keys())

    def __len__(self) -> int:
        return len(self.trajs)

    def take(
        self,
        size: Union[float, int],
        stratify: bool = True,
        shuffle: bool = True,
        random_state: Union[int, None] = None,
    ) -> Data:
        """Takes a subset of the dataset."""
        if isinstance(size, int):
            assert 0 < size < len(self), "size should be within 0 and len(self)"
            size /= len(self)

        ans, _ = self.split(
            size, stratify=stratify, shuffle=shuffle, random_state=random_state
        )
        return ans

    def cut(self, size: Union[float, int]):
        """
        Similar to split, but without shuffle, stratify, etc. Just slices the
        dataset into two parts.
        """
        if isinstance(size, float):
            assert 0 < size < 1, "size should be within 0 and 1 if float"
            size = int(len(self) * size)
        else:
            assert (
                0 < size < len(self)
            ), "size should be within 0 and the dataset size if int"

        left, right = self.trajs[:size], self.trajs[size:]
        left_labels, right_labels = self.labels[:size], self.labels[size:]
        left_d = Data(self.dataset, left, left_labels)
        right_d = Data(self.dataset, right, right_labels)
        return left_d, right_d

    def split(
        self,
        train_size: Union[float, int] = 0.8,
        stratify: bool = True,
        shuffle: bool = True,
        random_state: Union[int, None] = None,
    ) -> Tuple[Data, Data]:
        """
        Splits the dataset into train and test dataset slices.

        It uses the sklearn.model_selection.train_test_split function.

        Parameters
        ----------
        train_size : Union[float, int], optional
            The proportion of the dataset to include in the train split.
            If float, should be between 0.0 and 1.0, if int, represents the
            absolute number of train samples. By default 0.8.
        stratify : bool, optional
            If True, the split will be stratified according to the labels,
            by default True
        shuffle : bool, optional
            If True, the split will be shuffled, by default True
        random_state : Union[int, None], optional
            Random seed for reproducibility, by default None

        Returns
        -------
        Tuple[Data, Data]
            A tuple with the train and test Data objects.
        """
        if isinstance(train_size, int):
            assert (
                0 < train_size < len(self)
            ), "train_size should be within 0 and the dataset size if int"
            train_size /= len(self)
        else:
            assert 0 < train_size < 1, "train_size should be within 0 and 1 if float"

        x_train, x_test, y_train, y_test = train_test_split(
            self.trajs,
            self.labels,
            train_size=train_size,
            stratify=self.labels if stratify else None,
            random_state=random_state,
            shuffle=shuffle,
        )

        train_data = Data(self.dataset, x_train, y_train)
        test_data = Data(self.dataset, x_test, y_test)
        return train_data, test_data

    def map(self, func: Callable[[Trajectory, Any], Tuple[Trajectory, Any]]) -> Data:
        """
        Applies a function to each trajectory and label pair.

        Usefull to apply some preprocessing to the trajectories
        or the labels.

        Parameters
        ----------
        func : Callable[[Trajectory, Any], Tuple[Trajectory, Any]]
            Function to be applied to each trajectory and label pair.

        Returns
        -------
        Data
            A new Data object with the results of the function.
        """
        trajs, labels = [], []
        for traj, label in zip(self.trajs, self.labels):
            traj, label = func(traj, label)
            trajs.append(traj)
            labels.append(label)
        return Data(self.dataset, trajs, labels)

    def filter(self, func: Callable[[Trajectory, Any], bool]) -> Data:
        """
        Filters the dataset based on a function.

        Parameters
        ----------
        func : Callable[[Trajectory, Any], bool]
            Function to be applied to each trajectory and label pair.

        Returns
        -------
        Data
            A new Data object with the filtered trajectories and labels.
        """
        trajs, labels = [], []
        for traj, label in zip(self.trajs, self.labels):
            if func(traj, label):
                trajs.append(traj)
                labels.append(label)
        logging.info("Filtered %d of %d trajectories", len(trajs), len(self))
        return Data(self.dataset, trajs, labels)


class Dataset(Data):
    """Wraps the data with some general properties that describes a full dataset"""

    _last_tag = None

    def __init__(
        self,
        name: str,
        trajs: List[Trajectory],
        labels: List[Any],
        version: int = 0,
    ):
        self.name = name
        self.version = version
        self.trajs = trajs
        self.labels = labels
        super().__init__(self, trajs, labels)

    def __len__(self):
        return len(self.trajs)

    @staticmethod
    def _from_json(name: str, data: dict) -> Dataset:
        assert "trajs" in data, "trajs not found in dataset"
        assert "labels" in data, "labels not found in dataset"
        assert "version" in data, "version not found in dataset"
        trajs = [JSONSerializer.from_json(traj) for traj in data["trajs"]]
        return Dataset(
            name=name,
            trajs=trajs,
            labels=data["labels"],
            version=data["version"],
        )

    @staticmethod
    def _get_dataset_url(name: str) -> str:
        tag = Dataset._last_tag
        if tag is None:
            g_cmd = Git()
            output = g_cmd.ls_remote(REPO_URL, sort="-v:refname", tags=True)
            tag = output.split("\n")[0].split("/")[-1]
            Dataset._last_tag = tag
        assert tag is not None, "Could not find the last tag"
        return f"{REPO_URL}/releases/download/{tag}/{name}.zip"

    @staticmethod
    def _from_url(name: str, force: bool = False) -> Dataset:
        url = Dataset._get_dataset_url(name)
        raw_dir = _get_path(config.DS_DIR, name)
        yupi_data_file = raw_dir / f"{name}.json"

        if not force and yupi_data_file.exists():
            with open(yupi_data_file, "r", encoding="utf-8") as yupi_fd:
                data = json.load(yupi_fd)
                if "trajs" in data and "labels" in data and "version" in data:
                    return Dataset._from_json(name, data)
            logging.warning("Invalid dataset file, downloading again")

        download_dataset(url, name)
        return Dataset.from_file(yupi_data_file, name)

    @staticmethod
    def from_file(path: Union[Path, str], name: str) -> Dataset:
        """Loads a dataset from a file."""
        _path = path if isinstance(path, Path) else Path(path)
        with open(_path, "r", encoding="utf-8") as yupi_fd:
            data = json.load(yupi_fd)
            return Dataset._from_json(name, data)

    @staticmethod
    def geolife(redownload: bool = False) -> Dataset:
        """Loads the geolife dataset"""
        return Dataset._from_url("geolife", force=redownload)

    @staticmethod
    def animals(redownload: bool = False) -> Dataset:
        """Loads the animals dataset"""
        return Dataset._from_url("animals", force=redownload)

    @staticmethod
    def mnist_stroke(redownload: bool = False) -> Dataset:
        """Loads the mnist_stroke dataset"""
        return Dataset._from_url("mnist_stroke", force=redownload)

    @staticmethod
    def hurdat2(redownload: bool = False) -> Dataset:
        """Loads the hurdat2 dataset"""
        return Dataset._from_url("hurdat2", force=redownload)

    @staticmethod
    def cma_bst(redownload: bool = False) -> Dataset:
        """Loads the cma_bst dataset"""
        return Dataset._from_url("cma_bst", force=redownload)

    @staticmethod
    def uci_gotrack(redownload: bool = False) -> Dataset:
        """Loads the uci_gotrack dataset"""
        return Dataset._from_url("uci_gotrack", force=redownload)

    @staticmethod
    def uci_movement_libras(redownload: bool = False) -> Dataset:
        """Loads the uci_movement_libras dataset"""
        return Dataset._from_url("uci_movement_libras", force=redownload)

    @staticmethod
    def uci_pen_digits(redownload: bool = False) -> Dataset:
        """Loads the uci_pen_digits dataset"""
        return Dataset._from_url("uci_pen_digits", force=redownload)

    @staticmethod
    def uci_characters(redownload: bool = False) -> Dataset:
        """Loads the uci_characters dataset"""
        return Dataset._from_url("uci_characters", force=redownload)

    @staticmethod
    def stochastic_models(redownload: bool = False) -> Dataset:
        """Loads the stochastic models dataset"""
        return Dataset._from_url("stochastic_models", force=redownload)



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers


class TransformerBlock(layers.Layer):
    def __init__(self, head_size, num_heads, ff_dim, ff_dim2, rate=0.1):
        super().__init__()
        self.att = layers.MultiHeadAttention(num_heads=num_heads, key_dim=head_size)
        self.layernorm1 = layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = layers.Dropout(rate)
        self.dropout2 = layers.Dropout(rate)
        self.conv1 = layers.Conv1D(filters=ff_dim, kernel_size=1, activation="relu")
        self.conv2 = layers.Conv1D(filters=ff_dim2, kernel_size=1)
        self.supports_masking = True

    def call(self, inputs, training, mask=None):
        padding_mask = None
        if mask is not None:
            padding_mask = tf.cast(mask[:, tf.newaxis, tf.newaxis, :], dtype="int32")

        out_norm1 = self.layernorm1(inputs, training=training)
        out_att = self.att(
            out_norm1, out_norm1, training=training, attention_mask=padding_mask
        )
        out_drop1 = self.dropout1(out_att, training=training)
        res = out_drop1 + inputs
        out_norm2 = self.layernorm2(res, training=training)
        out_conv1 = self.conv1(out_norm2, training=training)
        out_drop2 = self.dropout2(out_conv1, training=training)
        out_conv2 = self.conv2(out_drop2, training=training)
        return out_conv2 + res


def build_model(
    n_classes,
    input_shape,
    head_size,
    num_heads,
    ff_dim,
    num_transformer_blocks,
    mlp_units,
    dropout=0.0,
    mlp_dropout=0.0,
    mask=None,
) -> keras.Model:
    inputs = keras.Input(shape=input_shape)
    _x = inputs
    if mask is not None:
        _x = layers.Masking(mask_value=mask)(_x)
    for _ in range(num_transformer_blocks):
        _x = TransformerBlock(
            head_size,
            num_heads,
            ff_dim,
            inputs.shape[-1],
            dropout,
        )(_x)

    _x = layers.GlobalAveragePooling2D(data_format="channels_first")(_x)
    for dim in mlp_units:
        _x = layers.Dense(dim, activation="relu")(_x)
        _x = layers.Dropout(mlp_dropout)(_x)
    outputs = layers.Dense(n_classes, activation="softmax")(_x)
    return keras.Model(inputs, outputs)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, List

from pactus.dataset import Data
from pactus.models.evaluation import Evaluation


def _mark_trained(func):
    def wrapper(self: Model, *args, **kwargs):
        logging.info("Training %s model", self.name)
        val = func(self, *args, **kwargs)
        self.trained = True
        return val

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


class Model(ABC):
    """Abstract class for models."""

    def __init__(self, name: str):
        self.name = name
        self.summary = {"name": name}
        self.trained = False

    @abstractmethod
    @_mark_trained
    def train(self, data: Data, cross_validation: int = 0):
        """Train the model using a given dataset"""

    @abstractmethod
    def predict(self, data: Data) -> List[Any]:
        """Predict the labels of a given set of trajectories"""

    def set_summary(self, **summary):
        """Set the summary of the model"""
        self.summary.update(summary)

    def save(self, path: str):
        """Save the model to a given path"""
        raise NotImplementedError

    def _predict(self, data: Data) -> List[Any]:
        # TODO: check mark trained decorator
        # if not self.trained:
        #     raise Exception("Model is not trained yet.")

        return self.predict(data)

    def evaluate(self, data: Data) -> Evaluation:
        """Evaluate the trained model"""

        logging.info("Evaluating the %s model", self.name)
        predictions = self._predict(data)
        return Evaluation(self.summary, data, predictions)



from string import Template
from typing import Any, Dict, List

from pactus.models.evaluation import Evaluation

LATEX_EVAL_ROW_TEMPLATE = Template(
    r"""	   \multirow{2}{*}{$model_name}
			   & ACC & $acc_vals \\
			   & F-Score & $f_score_vals\\
"""
)

LATEX_EVAL_TEMPLATE = Template(
    r"""
\begin{figure}[ht!]
\caption{$caption}
\vspace{2mm}
\centering
   \begin{tabular}{cr|$ds_cols_chars}
	   \toprule
	   \multirow{2}[4]{*}{\textbf{Model}} &
	   \multirow{2}[4]{*}{\textbf{Metrics}} &
	   \multicolumn{$ds_count}{c}{\textbf{Datasets}} \\
       & \cline{2-$c_line_top}
	   & & $ds_names \\
	   \midrule
$rows
	   \bottomrule
   \end{tabular}
\end{figure}
"""
)


class EvaluationComparison:
    def __init__(self, evals: List[Evaluation]):
        self.evals = evals

        self.evals_by_model: Dict[str, List[Evaluation]] = {}
        for evaluation in self.evals:
            model = evaluation.model_summary["name"]
            self.evals_by_model[model] = self.evals_by_model.get(model, []) + [
                evaluation
            ]

        self.evals_by_dataset: Dict[str, List[Evaluation]] = {}
        for evaluation in self.evals:
            ds_name = evaluation.dataset.name
            self.evals_by_dataset[ds_name] = self.evals_by_dataset.get(ds_name, []) + [
                evaluation
            ]

    def to_latex(self) -> str:
        rows = []
        for model, evaluations in self.evals_by_model.items():
            model_name = model.replace("_", " ").title()
            rows.append(
                LATEX_EVAL_ROW_TEMPLATE.substitute(
                    model_name=model_name,
                    acc_vals=" & ".join(
                        [f"{e.acc_overall:.2f}\\%" for e in evaluations]
                    ),
                    f_score_vals=" & ".join(
                        [f"{e.f1_score:.2f}\\%" for e in evaluations]
                    ),
                )
            )
        ds_names = [ds.replace("_", " ").title() for ds in self.evals_by_dataset]
        return LATEX_EVAL_TEMPLATE.substitute(
            caption="Evaluation Comparison",
            ds_cols_chars="c" * len(self.evals_by_dataset),
            c_line_top=len(self.evals_by_dataset) + 1,
            ds_names=" & ".join(ds_names),
            ds_count=len(self.evals_by_dataset),
            rows="".join(rows),
        )



from pactus.models.decision_tree_model import DecisionTreeModel
from pactus.models.evaluation import Evaluation
from pactus.models.evaluation_comparison import EvaluationComparison
from pactus.models.kneighbors_model import KNeighborsModel
from pactus.models.model import Model
from pactus.models.random_forest_model import RandomForestModel
from pactus.models.svm_model import SVMModel
from pactus.models.transformer_model import TransformerModel

__all__ = [
    "DecisionTreeModel",
    "Evaluation",
    "EvaluationComparison",
    "KNeighborsModel",
    "Model",
    "RandomForestModel",
    "SVMModel",
    "TransformerModel",
]



from typing import Any, List

from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from yupi import Trajectory

from pactus import featurizers
from pactus.dataset import Data
from pactus.models.model import Model

NAME = "SVM"


class SVMModel(Model):
    """Implementation of a Support Vector Machine Classifier."""

    def __init__(self, featurizer: featurizers.Featurizer, **kwargs):
        super().__init__(NAME)
        self.featurizer = featurizer
        self.model = SVC(**kwargs)
        self.grid: GridSearchCV
        self.set_summary(**kwargs)

    def train(self, data: Data, cross_validation: int = 0):
        self.set_summary(cross_validation=cross_validation)
        x_data = self.featurizer.featurize(data.trajs)
        self.grid = GridSearchCV(self.model, {}, cv=cross_validation, verbose=3)
        self.grid.fit(x_data, data.labels)

    def predict(self, data: Data) -> List[Any]:
        x_data = self.featurizer.featurize(data.trajs)
        return self.grid.predict(x_data)

    def predict_single(self, traj: Trajectory) -> Any:
        """Predicts the label of a single trajectory."""
        return self.grid.predict([traj])[0]



import gzip
import logging
import tarfile
import zipfile
from pathlib import Path

import requests
from requests import Response

import pactus.config as cfg


def _get_path(path: str, *args) -> Path:
    return Path(path.format(*args))


def _get_progress_log(part, total):
    # If the total is unknown, just return the part
    if total == -1:
        return f"Downloaded: {part / 1024 ** 2:.2f} MB"

    passed = "=" * int(cfg.PROGRESS_BAR_LENGTH * part / total)
    rest = " " * (cfg.PROGRESS_BAR_LENGTH - len(passed))
    p_bar = f"[{passed}{rest}] {part * 100/total:.2f}%"
    if part == total:
        p_bar += "\n"
    return p_bar


def _create_dataset_path(dataset_name: str) -> Path:
    logging.info("Creating dataset folder for %s", dataset_name)
    dataset_path = _get_path(cfg.DS_DIR, dataset_name)
    dataset_path.mkdir(parents=True, exist_ok=True)
    return dataset_path


def _start_download(url: str, dataset_name: str) -> Response:
    logging.info("Downloading %s dataset", dataset_name)
    response = requests.get(url, allow_redirects=True, stream=True)
    if response.status_code != 200:
        raise RuntimeError(f"Failed to download dataset {dataset_name}")
    return response


def _download_until_finish(url: str, response: Response, dataset_path: Path) -> Path:
    data_length = int(response.headers.get("content-length", -1))
    size_mb_msg = (
        f"    Size: {data_length / 1024 ** 2:.2f} MB" if data_length != -1 else ""
    )
    dataset_file_path = dataset_path / url.split("/")[-1]
    with open(dataset_file_path, "wb") as ds_file:
        downloaded = 0
        for chunk in response.iter_content(chunk_size=cfg.DOWNLOAD_CHUNCK_SIZE):
            if chunk:
                ds_file.write(chunk)
                downloaded += len(chunk)
                print(
                    _get_progress_log(downloaded, data_length) + size_mb_msg, end="\r"
                )
    return dataset_file_path


def _download(url: str, dataset_name: str, dataset_path: Path) -> Path:
    # Make the download request
    response = _start_download(url, dataset_name)

    # Chekc if the dataset is already downloaded
    dataset_file_path = dataset_path / url.split("/")[-1]
    if dataset_file_path.exists():
        size = -1
        if "content-length" in response.headers:
            size = int(response.headers["content-length"])
            if size == dataset_file_path.stat().st_size:
                logging.info("Dataset already downloaded")
                return dataset_file_path

        msg = (
            "There is a downloaded dataset with the same name, but the download "
            "size is unknown."
        )

        if size != -1:
            msg = (
                "It seems that the dataset is already downloaded, but the size is "
                f"different.\n"
                f"    Found: {dataset_file_path.stat().st_size / 1024 ** 2:.2f} MB\n"
                f"    Expected: {size / 1024 ** 2:.2f} MB\n"
            )
        logging.warning(msg)
        ans = input("Do you want to overwrite it? [y/n]: ")
        if ans.lower() != "y":
            return dataset_file_path

    # Download the dataset to a zip file
    return _download_until_finish(url, response, dataset_path)


def _uncompress(dataset_name: str, dataset_file_path: Path, dataset_path: Path):
    logging.info("Extracting %s dataset", dataset_name)

    str_path = str(dataset_file_path)
    if str_path.endswith(".zip"):
        with zipfile.ZipFile(dataset_file_path, "r") as zip_ref:
            zip_ref.extractall(dataset_path)
    elif str_path.endswith(".tar.gz"):
        with tarfile.open(dataset_file_path, "r") as tar_ref:
            tar_ref.extractall(dataset_path)
    elif str_path.endswith(".gz"):
        with gzip.open(dataset_file_path, "rb") as gz_ref:
            with open(str(dataset_file_path.with_suffix("")), "wb") as uncompressed:
                uncompressed.write(gz_ref.read())


def download_dataset(url: str, dataset_name: str) -> None:
    """Downloads a dataset from a url."""

    if not url.endswith(".zip") and not url.endswith(".gz"):
        raise ValueError("The dataset must be a zip or .gz file")

    # Create the dataset folder if it doesn't exist
    dataset_path = _create_dataset_path(dataset_name)

    # Download the compressed version of the dataset
    dataset_file_path = _download(url, dataset_name, dataset_path)

    # Extract the dataset
    _uncompress(dataset_name, dataset_file_path, dataset_path)
